from enum import Enum
from datetime import datetime
import math
import asyncio
import json

import aiohttp

from ..gametypes import GameType

RAW_ACHIEVEMENTS_URL = \
    "https://raw.githubusercontent.com/HypixelDev" \
    "/PublicAPI/master/Documentation/misc/Achievements.json"

__all__ = [
    "OneTimeAchievement", "TieredAchievement", "AchievementTier",
    "PixelAchievements", "PlayerAchievements", "PlayerRank", "Player"
]


class OneTimeAchievement:
    def __init__(self, data: tuple):
        self.game = data[0]
        self.db_name = data[1]
        self.points = data[2]["points"]
        self.name = data[2]["name"]
        self.description = data[2]["description"]

    def __repr__(self):
        return "<OneTimeAchievement game={0.game} db_name={0.db_name} "\
               "name={0.name} points={0.points} description={0.description}"\
               "".format(self)


class TieredAchievement:
    def __init__(self, data: tuple):
        self.game = data[0]
        self.db_name = data[1]
        self.player_value = data[3]
        self.name = data[2]["name"]
        self.description = data[2]["description"]
        completed_tiers = []
        unfinished_tiers = []
        for tier in data[2]["tiers"]:
            if tier["amount"] <= self.player_value:
                completed_tiers.append(AchievementTier(**tier))
            else:
                unfinished_tiers.append(AchievementTier(**tier))
        self.completed_tiers = completed_tiers
        self.unfinished_tiers = unfinished_tiers


class AchievementTier:
    def __init__(self, tier: int, points: int, amount: int):
        self.tier = tier
        self.points = points
        self.amount = amount


class PixelAchievements:
    achievements = None

    def __init__(self):
        loop = asyncio.get_event_loop()
        loop.create_task(self.get_all_achievements())
        self.achievements = None

    async def get_all_achievements(self):
        async with aiohttp.ClientSession() as session:
            async with session.get(RAW_ACHIEVEMENTS_URL) as r:
                if r.status == 200:
                    try:
                        data = json.loads(await r.text())
                    except aiohttp.ClientResponseError as e:
                        pass
                    else:
                        self.achievements = data["achievements"]

    def get_one_time_achievement(self, name: str):
        game, achievement_name = name.split("_", 1)
        if not achievement_name.upper() in self.achievements[game]["one_time"]:
            return None
        raw_achievement = \
            self.achievements[game]["one_time"][achievement_name.upper()]
        return OneTimeAchievement((game, achievement_name, raw_achievement))

    def get_tiered_achievement(self, name: str, player_value: int):
        game, achievement_name = name.split("_", 1)
        if not achievement_name.upper() in self.achievements[game]["tiered"]:
            return None
        raw_achv = self.achievements[game]["tiered"][achievement_name.upper()]
        return TieredAchievement(
            (game, achievement_name, raw_achv, player_value)
        )


class PlayerAchievements:

    def __init__(self, data, all_achievements: PixelAchievements):
        self.all_achievements = all_achievements
        one_time = []
        tiered = []
        if "achievementsOneTime" in data:
            for item in data["achievementsOneTime"]:
                one_time.append(
                    self.all_achievements.get_one_time_achievement(item)
                )
        if "achievements" in data:
            for k, v in data["achievements"].items():
                tiered.append(
                    self.all_achievements.get_tiered_achievement(k, v)
                )
        self.one_time = one_time
        self.tiered = tiered


class PlayerRank(Enum):
    """
    An enum representing player ranks
    """
    ADMIN = "Admin"
    MODERATOR = "Moderator"
    HELPER = "Helper"
    JR_HELPER = "Jr. Helper"
    YOUTUBER = "Youtuber"
    BUILD_TEAM = "Build Team"
    SUPERSTAR = "MVP++"
    MVP_PLUS = "MVP+"
    MVP = "MVP"
    VIP_PLUS = "VIP+"
    VIP = "VIP"
    NONE = ""

    def __init__(self, pretty_name: str):
        self.pretty_name = pretty_name

    @classmethod
    def from_player_data(cls, data: dict):
        """
        Gets the player's rank from their player data
        """
        if "rank" in data:
            return getattr(cls, data["rank"])
        elif data.get("buildTeam") is True:
            return cls.BUILD_TEAM
        elif data.get("monthlyPackageRank") == "SUPERSTAR":
            return cls.SUPERSTAR
        elif "newPackageRank" in data:
            return getattr(cls, data["newPackageRank"])
        elif "packageRank" in data:
            return getattr(cls, data["packageRank"])
        else:
            return cls.NONE


class Player:
    def __init__(self, _id: str, displayname: str, first_login: int,
                 last_login: int, karma: int, network_exp: float,
                 rank: PlayerRank, last_logout: int,
                 achievements: PlayerAchievements, stats: list,
                 most_recent_game_type: GameType):
        self._id = _id
        self.displayname = displayname
        self.first_login = datetime.utcfromtimestamp(first_login/1000)
        self.last_login = datetime.utcfromtimestamp(last_login/1000)
        self.karma = karma
        self.network_exp = network_exp
        self.rank = rank
        self.last_logout = datetime.utcfromtimestamp(last_logout/1000)
        self.achievements = achievements
        self.stats = stats
        self.most_recent_game_type = most_recent_game_type

    def network_level(self) -> float:
        """
        Calculate the player's network level

        Returns
        -------
        float
            The player's network level
        """
        base = 10000
        growth = 2500

        reverse_pq_prefix = -(base - 0.5 * growth) / growth
        reverse_const = math.pow(reverse_pq_prefix, 2)
        growth_divides_2 = 2 / growth

        return 1.0 if self.network_exp < 0 \
            else 1 + reverse_pq_prefix + math.sqrt(
                reverse_const + growth_divides_2 * self.network_exp
            )
    
    def online(self) -> bool:
        """
        Show whether the player is online or not
        """
        return self.last_logout < self.last_login



from enum import Enum


class GameType(Enum):
    """
    An enum of game types
    """

    # This is formatted the way it is for readability as if it were a table
    # much like the one at
    # https://github.com/HypixelDev/PublicAPI/blob/master/Documentation/misc/GameType.md
    # Format:
    # Type name       ID  DB Name          Clean Name
    QUAKECRAFT     = (2,  "Quake",         "Quake")
    WALLS          = (3,  "Walls",         "Walls")
    PAINTBALL      = (4,  "Paintball",     "Paintball")
    SURVIVAL_GAMES = (5,  "HungerGames",   "Blitz Survival Games")
    TNTGAMES       = (6,  "TNTGames",      "TNT Games")
    VAMPIREZ       = (7,  "VampireZ",      "VampireZ")
    WALLS3         = (13, "Walls3",        "Mega Walls")
    ARCADE         = (14, "Arcade",        "Arcade")
    ARENA          = (17, "Arena",         "Arena")
    UHC            = (20, "UHC",           "UHC Champions")
    MCGO           = (21, "MCGO",          "Cops and Crims")
    BATTLEGROUND   = (23, "Battleground",  "Warlords")
    SUPER_SMASH    = (24, "SuperSmash",    "Smash Heroes")
    GINGERBREAD    = (25, "GingerBread",   "Turbo Kart Racers")
    HOUSING        = (26, "Housing",       "Housing")
    SKYWARS        = (51, "SkyWars",       "SkyWars")
    TRUE_COMBAT    = (52, "TrueCombat",    "Crazy Walls")
    SPEED_UHC      = (54, "SpeedUHC",      "Speed UHC")
    SKYCLASH       = (55, "SkyClash",      "SkyClash")
    LEGACY         = (56, "Legacy",        "Classic Games")
    PROTOTYPE      = (57, "Prototype",     "Prototype")
    BEDWARS        = (58, "Bedwars",       "Bed Wars")
    MURDER_MYSTERY = (59, "MurderMystery", "Murder Mystery")
    BUILD_BATTLE   = (60, "BuildBattle",   "Build Battle")
    DUELS          = (61, "Duels",         "Duels")
    SKYBLOCK       = (None, None, None)

    def __init__(self, _id: int, db_name: str, name: str):
        self.id = _id
        self.db_name = db_name
        self.clean_name = name

    @classmethod
    def from_id(cls, _id: int):
        for gt, data in cls.__members__.items():
            if _id == data.id:
                return data
        return None

    @classmethod
    def from_db_name(cls, db_name: str):
        for gt, data in cls.__members__.items():
            if db_name == data.db_name:
                return data
        return None

    @classmethod
    def from_clean_name(cls, clean_name: str):
        for gt, data in cls.__members__.items():
            if clean_name == data.clean_name:
                return data
        return None



from typing import List
from ..utils import get_player_name
from ..gametypes import GameType

__all__ = ["Leaderboard", "LeaderboardMember"]


class LeaderboardMember:
    def __init__(self, uuid: str):
        self.uuid = uuid

    async def name(self):
        """
        Get the username of a leaderboard member
        """
        return await get_player_name(self.uuid)


class Leaderboard:
    def __init__(self, path: str, prefix: str, count: int,
                 leaders: List[LeaderboardMember],
                 title: str, game_type: GameType):
        self.path = path
        self.prefix = prefix
        self.count = count
        self.leaders = leaders
        self.title = title
        self.game_type = game_type



from datetime import datetime

from ..utils import get_player_name


class Friend:

    def __init__(self, _id: str, sender_uuid: str,
                 receiver_uuid: str, started: int):
        self._id = _id
        self.sender_uuid = sender_uuid
        self.receiver_uuid = receiver_uuid
        self.started = datetime.utcfromtimestamp(started / 1000)

    def __repr__(self):
        return "<Friend _id={0._id} sender_uuid={0.sender_uuid} " \
               "receiver_uuid={0.receiver_uuid}>".format(self)

    async def sender_name(self):
        """
        Get the username of the player who sent the friend request
        """
        return await get_player_name(self.sender_uuid)

    async def receiver_name(self):
        """
        Get the username of the player who received the friend request
        """
        return await get_player_name(self.receiver_uuid)



from .boosters import Booster
from .friends import Friend
from .guilds import GuildMember, GuildTag, Guild
from .leaderboards import Leaderboard, LeaderboardMember
from .players import Player, PlayerAchievements, PlayerRank, \
    PixelAchievements, TieredAchievement, AchievementTier, OneTimeAchievement
from .sessions import PlayerSession
from .stats import *



from datetime import datetime
from typing import List
from enum import Enum

from ..utils import get_player_name
from ..gametypes import GameType

__all__ = ["Guild", "GuildMember", "GuildRank", "GuildTag"]

"""
class GuildPermissions(Enum):
    # Format:
    # Some name         ID   Name                    Description
    MODIFY_GUILD_NAME = (0,  "Modify Guild Name",    "Change the guild's name.")
    MODIFY_GUILD_MOTD = (1,  "Modify Guild MOTD",    "Change the guild's message of the day.")
    MODIFY_GUILD_TAG  = (2,  "Modify Guild Tag",     "Change the guild's tag.")
    MANAGE_RANKS      = (3,  "Change Ranks",         "Promote or demote members (up to their own rank).")
    CHANGE_VISIBILITY = (4,  "Guild Finder options", "Change how the guild is shown in the Guild Finder, if at all.")
    OFFICER_CHAT      = (5,  "Officer Chat",         "Send and receive messages in the officer chat.")
    KICK_MEMBERS      = (6,  "Kick Members",         "Kick members from the guild.")
    MUTE_MEMBERS      = (7,  "Mute Members",         "Mute guild members.")
    INVITE_MEMBERS    = (8,  "Invite Members",       "Invite members to the guild.")
    VIEW_AUDIT_LOG    = (9,  "View Audit Log",       "View the audit log.")
    VIEW_STATS        = (10, "View Stats",           "View a guild member's stats.")
    START_PARTY       = (11, "Guild Party",          "Start a guild party")
    CHANGE_SETTINGS   = (12, "Guild Settings",       "Change the guild settings.")
    CHANGE_DISCORD    = (13, "Change Guild Discord", "Change the guild's Discord join link.")

    def __init__(self, _id: int, name: str, description: str):
        self.id = _id
        self.perm_name = name
        self.description = description
    
    @classmethod
    def from_id(cls, _id: int):
        for _, data in cls.__members__.items():
            if _id == data.id:
                return data
        return None


class GuildBanner:
    def __init__(self, base, patterns):
        self.base = base
        self.patterns = []
        for pattern in patterns:
            self.patterns.append(
                BannerPattern(
                    pattern["Pattern"], pattern["Color"]
                )
            )


class BannerPattern:
    def __init__(self, pattern, color):
        self.pattern = pattern
        self.color = color

"""

class GuildRank:
    def __init__(self, rank: dict):
        self.name = rank["name"]
        # self.permissions = [GuildPermissions.from_id(p) for p in rank["permissions"]]
        self.default = rank["default"]
        self.tag = rank["tag"]
        self.created = datetime.utcfromtimestamp(rank["created"]/1000)
        self.priority = rank["priority"]


class GuildMember:
    """
    A guild member
    """
    
    def __init__(self, member: dict):
        self.uuid = member["uuid"]
        self.rank = member["rank"]
        self.joined = datetime.utcfromtimestamp(member["joined"]/1000)

    async def name(self):
        """
        Gets the current username for this guild member

        Returns
        -------
        str
           The guild member's username
        """
        return await get_player_name(self.uuid)


class GuildTag:
    """
    A guild tag
    """
    def __init__(self, name: str, color: str):
        self.name = name
        self.color = color


class Guild:
    """
    A guild
    """
    def __init__(self, _id: str, created: int, exp: int, discord: str, joinable: bool,
                 description: str, members: List[GuildMember], name: str, tag: GuildTag,
                 vip_count: int, mvp_count: int, ranks: List[GuildRank],
                 preferred_games: List[GameType], chat_throttle: int, is_listed: bool,
                 chat_muted_until: datetime):
        self._id = _id
        self.created = datetime.utcfromtimestamp(created/1000)
        self.exp = exp
        self.discord = discord
        self.joinable = joinable
        self.description = description
        self.members = members
        self.name = name
        self.tag = tag
        # self.banner = banner
        self.vip_count = vip_count
        self.mvp_count = mvp_count
        self.ranks = ranks
        self.preferred_games = preferred_games
        self.chat_throttle = chat_throttle
        self.is_listed = is_listed
        self.chat_muted_until = chat_muted_until



class PixelException(Exception):
    pass


class InvalidKeyException(PixelException):
    pass


class GuildNotFound(PixelException):
    pass


class PlayerNotInGuild(PixelException):
    pass


class PlayerNotFound(PixelException):
    pass


class NoSessionForPlayer(PixelException):
    pass



from ..gametypes import GameType


class PlayerSession:
    def __init__(self, data: dict):
        self.game_type = getattr(GameType, data["gameType"])
        self.server = data["server"]
        self.players = data["players"]



import re

import aiohttp


async def get_player_name(uuid: str, session=None):
    session = aiohttp.ClientSession() if session is None else session
    url = "https://api.mojang.com/user/profiles/{}/names".format(clean_uuid(uuid))

    async with session.get(url) as r:
        data = await r.json()
    return data[-1]["name"]


async def get_player_uuid(name: str, session=None):
    session = aiohttp.ClientSession() if session is None else session
    headers = {
        "Content-Type": "application/json"
    }
    r_data = [name]
    url = "https://api.mojang.com/profiles/minecraft"
    async with session.post(url, json=r_data, headers=headers) as r:
        data = await r.json()
    if data:
        return data[0]["id"]
    else:
        return None


def clean_uuid(uuid: str) -> str:
    return re.sub("-", "", uuid)



from ..gametypes import GameType

__all__ = [
    "PlayerStats", "QuakeStats", "WallsStats", "PaintballStats",
    "HungerGamesStats", "TNTGamesStats", "VampireZStats", 
    "Walls3Stats", "ArcadeStats", "ArenaStats", "UHCStats", 
    "MCGOStats", "BattlegroundStats", "SuperSmashStats", 
    "GingerBreadStats", "HousingStats", "SkyWarsStats", 
    "TrueCombatStats", "SpeedUHCStats", "SkyClashStats", 
    "LegacyStats", "PrototypeStats", "BedwarsStats", 
    "MurderMysteryStats", "BuildBattleStats", "DuelsStats"
]

class PlayerStats:
    def __init__(self, data: dict):
        for k, v in data.items():
            setattr(self, k, v)


class QuakeStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.QUAKECRAFT
        super().__init__(data)


class WallsStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.WALLS
        super().__init__(data)


class PaintballStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.PAINTBALL
        super().__init__(data)


class HungerGamesStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.SURVIVAL_GAMES
        super().__init__(data)


class TNTGamesStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.TNTGAMES
        super().__init__(data)


class VampireZStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.VAMPIREZ
        super().__init__(data)


class Walls3Stats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.WALLS3
        super().__init__(data)


class ArcadeStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.ARCADE
        super().__init__(data)


class ArenaStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.ARENA
        super().__init__(data)


class UHCStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.UHC
        super().__init__(data)


class MCGOStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.MCGO
        super().__init__(data)


class BattlegroundStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.BATTLEGROUND
        super().__init__(data)


class SuperSmashStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.SUPER_SMASH
        super().__init__(data)


class GingerBreadStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.GINGERBREAD
        super().__init__(data)


class HousingStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.HOUSING
        super().__init__(data)


class SkyWarsStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.SKYWARS
        super().__init__(data)


class TrueCombatStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.TRUE_COMBAT
        super().__init__(data)


class SpeedUHCStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.SPEED_UHC
        super().__init__(data)


class SkyClashStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.SKYCLASH
        super().__init__(data)


class LegacyStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.LEGACY
        super().__init__(data)


class PrototypeStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.PROTOTYPE
        super().__init__(data)


class BedwarsStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.BEDWARS
        super().__init__(data)

    def level(self) -> float:
        """
        Calculate the player's Bedwars level

        Returns
        -------
        float:
            The player's Bedwars level
        """
        exp_per_prestige = 489000
        levels_per_prestige = 100

        exp = getattr(self, "Experience", 0.0)
        prestige = int(exp/exp_per_prestige)
        exp = exp % exp_per_prestige

        if prestige > 5:
            over = prestige % 5
            exp += over * exp_per_prestige
            prestige -= over

        if exp < 500:
            return 0 + prestige * levels_per_prestige
        elif exp < 1500:
            return 1 + prestige * levels_per_prestige
        elif exp < 3500:
            return 2 + prestige * levels_per_prestige
        elif exp < 5500:
            return 3 + prestige * levels_per_prestige
        elif exp < 9000:
            return 4 + prestige * levels_per_prestige

        exp -= 9000
        return (exp / 5000 + 4) + (prestige * levels_per_prestige)


class MurderMysteryStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.MURDER_MYSTERY
        super().__init__(data)


class BuildBattleStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.BUILD_BATTLE
        super().__init__(data)


class DuelsStats(PlayerStats):
    def __init__(self, data: dict):
        self.game = GameType.DUELS
        super().__init__(data)



from datetime import datetime

from aiopixel import GameType
from ..utils import get_player_name


class Booster:
    """
    A booster
    """

    def __init__(self, _id: str, purchaser_uuid: str, amount: float,
                 original_length: int, length: int, game_type: int,
                 activated_at: int, stacked: list=None):
        self._id = id
        self.purchaser_uuid = purchaser_uuid
        self.amount = amount
        self.original_length = original_length
        self.length = length
        self.game_type = GameType.from_id(game_type)
        self.activated_at = datetime.utcfromtimestamp(activated_at / 1000)
        self.stacked = stacked

    def __repr__(self):
        return "<Booster id={0._id} purchaser={0.purchaser} " \
               "length={0.length} game_type={0.game_type.id}>".format(self)

    async def purchaser_name(self):
        """
        Get the purchaser's name
        """
        return await get_player_name(self.purchaser_uuid)



from distutils.core import setup
from setuptools import find_packages


def get_requirements():
    with open("requirements.txt") as f:
        requirements = f.read().splitlines()
    return requirements


setup(
    name="aiopixel",
    version="0.2.4",
    packages=find_packages(include=["aiopixel", "aiopixel.*"]),
    license="GPLv3",
    install_requires=get_requirements(),
    python_requires=">=3.5",
    url="https://github.com/palmtree5/aiopixel",
    author="palmtree5",
    author_email="",
)



from typing import List

import aiohttp

from .exceptions import (GuildNotFound, InvalidKeyException,
                         NoSessionForPlayer, PixelException, PlayerNotFound,
                         PlayerNotInGuild)
from .gametypes import GameType
from .models import stats
from .models.boosters import Booster
from .models.friends import Friend
from .models.guilds import Guild, GuildMember, GuildTag, GuildRank
from .models.leaderboards import Leaderboard, LeaderboardMember
from .models.players import (PixelAchievements, Player, PlayerAchievements,
                             PlayerRank)
from .models.sessions import PlayerSession
from .utils import clean_uuid, get_player_uuid

BASE_API_URL = "https://api.hypixel.net{}"

__all__ = ["PixelClient"]


class PixelClient:
    """
    Asynchronus client for accessing Hypixel's API
    """
    def __init__(self, api_key: str):
        self._api_key = api_key
        self._session = aiohttp.ClientSession()
        self._achievements = PixelAchievements()

    async def _request(self, route: str, method: str, params: dict=None):
        async with self._session.request(method, BASE_API_URL.format(route), params=params) as r:
            if not r.status == 200:
                raise PixelException(
                    "An error occurred while making the request.\n"
                    "Status code: {0.status}.\n"
                    "Reason: {0.reason}".format(r)
                )
            data = await r.json()
            if not data["success"]:
                if data["cause"] == "Invalid API key!":
                    raise InvalidKeyException(data["cause"])
                else:
                    raise PixelException("Error: {}".format(data["cause"]))
            else:
                return data

    async def boosters(self) -> List[Booster]:
        """
        Get a list of all boosters

        Returns
        -------
        `list` of `Booster`
            A list of all boosters in the queue
        
        Raises
        ------
        InvalidKeyException
            If the client's API key is not valid
        
        PixelException
            If an error occurs while making the request or is present in the response json
        """
        data = await self._request(
            "/boosters", "GET", params={"key": self._api_key}
        )
        raw_boosters = data["boosters"]
        boosters = []
        for r_b in raw_boosters:
            boosters.append(Booster(
                r_b["_id"], r_b["purchaserUuid"], r_b["amount"],
                r_b["originalLength"], r_b["length"], r_b["gameType"],
                r_b["dateActivated"], stacked=r_b.get("stacked", None)
            ))
        return boosters

    async def find_guild_by_name(self, name: str) -> str:
        """
        Get the guild id for a guild with the specified name

        Parameters
        ----------
        name: str
            The guild name

        Returns
        -------
        str
            The guild id
        
        Raises
        ------
        InvalidKeyException
            If the client's API key is not valid
        
        PixelException
            If an error occurs while making the request or is present in the response json
        
        GuildNotFound
            If no guild is found
        """
        data = await self._request(
            "/findGuild", "GET", 
            params={"key": self._api_key, "byName": name}
        )
        if data["guild"] is None:
            raise GuildNotFound()
        return data["guild"]

    async def find_guild_by_uuid(self, uuid: str) -> str:
        """
        Get the guild id for the specified player's guild

        Parameters
        ----------
        uuid: str
            The player's uuid

        Returns
        -------
        str
            The guild id
        
        Raises
        ------
        InvalidKeyException
            If the client's API key is not valid
        
        PixelException
            If an error occurs while making the request or is present in the response json
        
        PlayerNotInGuild
            If no guild is found
        """
        data = await self._request(
            "/findGuild", "GET", 
            params={"key": self._api_key, "byUuid": clean_uuid(uuid)}
        )
        if data["guild"] is None:
            raise PlayerNotInGuild()
        return data["guild"]

    async def friends(self, uuid: str) -> List[Friend]:
        """
        Get friends of a player

        Parameters
        ----------
        uuid: str
            The uuid of the player to get friends of
        
        Returns
        -------
        `list` of `Friend`
            A list of objects representing the player's friendships
        
        Raises
        ------
        InvalidKeyException
            If the client's API key is not valid
        
        PixelException
            If an error occurs while making the request or if an error is returned in the response json
        """
        data = await self._request(
            "/friends", "GET", 
            params={"key": self._api_key, "uuid": clean_uuid(uuid)}
        )

        friends_list = []
        for raw_friend in data["records"]:
            friends_list.append(
                Friend(
                    raw_friend["_id"],
                    raw_friend["uuidSender"],
                    raw_friend["uuidReceiver"],
                    raw_friend["started"]
                )
            )
        return friends_list

    async def guild(self, guild_id: str) -> Guild:
        """
        Get a guild

        Parameters
        ----------
        guild_id: str
            The id of the guild to look up
        
        Returns
        -------
        Guild
            The requested guild
        
        Raises
        ------
        InvalidKeyException
            If the client's API key is not valid
        
        PixelException
            If an error occurs while making the request or is present in the response json
        
        GuildNotFound
            If no guild is found
        """
        data = await self._request(
            "/guild", "GET", 
            params={"key": self._api_key, "id": guild_id}
        )

        if data["guild"] is None:
            raise GuildNotFound()
        guild = data["guild"]
        members = []
        for m in guild["members"]:
            members.append(GuildMember(m))

        tag = GuildTag(guild.get("tag", None), guild.get("tagColor", None))
        guild_joinable = guild.get("joinable", False)
        vip_count = guild.get("vipCount", 0)
        mvp_count = guild.get("mvpCount", 0)
        discord = guild.get("discord", None)
        description = guild.get("description", None)
        ranks = guild.get("ranks", None)
        if ranks:
            ranks = [GuildRank(r) for r in ranks]
        preferred_games = guild.get("preferredGames", None)
        if preferred_games:
            preferred_games = [getattr(GameType, game) for game in preferred_games]
        # banner = GuildBanner(
        #     guild["banner"]["Base"], guild["banner"]["Patterns"]
        # )
        return Guild(
            guild["_id"], guild["created"], guild["exp"], discord, guild_joinable,
            description, members, guild["name"], tag,
            vip_count, mvp_count, ranks, preferred_games, guild.get("chatThrottle", 0),
            guild.get("publiclyListed", False), guild.get("chatMute", 0)
        )

    async def leaderboards(self) -> List[Leaderboard]:
        """
        Get all leaderboards
        
        Returns
        -------
        `list` of `Leaderboard`
            A list of all leaderboards
        
        Raises
        ------
        InvalidKeyException
            If the client's API key is not valid
        
        PixelException
            If an error occurs while making the request or is present in the response json
        """
        data = await self._request(
            "/leaderboards", "GET", 
            params={"key": self._api_key}
        )

        raw_leaderboards = data["leaderboards"]
        leaderboard_list = []
        for game, lbs in raw_leaderboards.items():
            game_type = getattr(GameType, game)
            for lb in lbs:
                leaders = []
                for l in lb["leaders"]:
                    leaders.append(LeaderboardMember(l))
                leaderboard_list.append(
                    Leaderboard(
                        lb["path"], lb["prefix"], lb["count"],
                        leaders, lb["title"], game_type
                    )
                )
        return leaderboard_list

    async def player_from_name(self, name: str) -> Player:
        """
        Gets the player represented by the specified name

        Parameters
        ----------
        name: str
            The name of the player
        
        Returns
        -------
        Player
            The specified player
        
        Raises
        ------
        InvalidKeyException
            If the client's API key is not valid
        
        PixelException
            If an error occurs while making the request or is present in the response json
        
        PlayerNotFound
            If the player doesn't exist or has never logged into the server
        """
        uuid = await get_player_uuid(name, self._session)
        if uuid is None:
            raise PlayerNotFound()
        return await self.player_from_uuid(uuid)

    async def player_from_uuid(self, uuid: str) -> Player:
        """
        Gets the player represented by the specified uuid

        Parameters
        ----------
        uuid: str
            The uuid of the player
        
        Returns
        -------
        Player
            The specified player
        
        Raises
        ------
        InvalidKeyException
            If the client's API key is not valid
        
        PixelException
            If an error occurs while making the request or is present in the response json
        
        PlayerNotFound
            If the player doesn't exist or has never logged into the server
        """
        data = await self._request(
            "/player", "GET", 
            params={"key": self._api_key, "uuid": uuid}
        )

        if not bool(data["player"]):
            raise PlayerNotFound()
        player_data = data["player"]
        player_rank = PlayerRank.from_player_data(player_data)
        player_achievements = PlayerAchievements(player_data, self._achievements)
        most_recent_game_type = getattr(GameType, player_data["mostRecentGameType"])
        player_stats = []
        for k, v in player_data["stats"].items():
            player_stats.append(getattr(stats, k + "Stats")(v))
        if "housingMeta" in player_data:
            player_stats.append(stats.HousingStats(player_data["housingMeta"]))
        return Player(
            player_data["_id"], player_data["displayname"], player_data["firstLogin"],
            player_data["lastLogin"], player_data["karma"], player_data["networkExp"],
            player_rank, player_data["lastLogout"], player_achievements,
            player_stats, most_recent_game_type
        )
    
    async def session(self, uuid: str) -> PlayerSession:
        """
        Get a player's current session

        Parameters
        ----------
        uuid: str
            The player to get a session for
        
        Returns
        -------
        PlayerSession
            The player's current session
        
        Raises
        ------
        InvalidKeyException
            If the client's API key is not valid
        
        PixelException
            If an error occurs while making the request or is present in the response json
        
        NoSessionForPlayer
            If the specified player does not have a session
        """
        data = await self._request(
            "/session", "GET", 
            params={"key": self._api_key, "uuid": uuid}
        )

        if not data["session"]:
            raise NoSessionForPlayer()
        else:
            return PlayerSession(data["session"])
